#include <allegro.h>
#include <stdio.h>
#include <math.h>
#include "main.h"
#include "game.h"

PLAYER players[4];
int nplayers;

void RedrawLevel(LEVEL *, BITMAP *, BITMAP *);
void Gravity(LEVEL *);
void UnlockMulti();
int MoveObjects(LEVEL *);
int InputCheck();
int AllTrollsDead(LEVEL *);
int CanMoveHere(LEVEL *, int, int);

void Storyline()
{
	LEVEL *lvl;
	int s;

	clear_bitmap(screenbuf);
	textout(screenbuf, font, "You are Potted the Hairy, and it is your job to stop the evil Lord", 8, 8, makecol(255, 255, 255));
	textout(screenbuf, font, "what's-his-name. Just don't ask what his name is.", 8, 16, makecol(255, 255, 255));
	textout(screenbuf, font, "Anyways, he (he's a troll, by the way) has replicated himself!", 8, 24, makecol(255, 255, 255));
	textout(screenbuf, font, "You must defeat all of him!", 8, 32, makecol(255, 255, 255));
	BLITBUF();
	clear_keybuf();
	readkey();

	lvl = LevelLoad("level0.lvl");
	s = RunLevel(lvl);
	LevelFree(lvl);
	if(s == 1)
	{
		UnlockMulti();

		clear_bitmap(screenbuf);
		textout(screenbuf, font, "Great job! You've won! Now you can play multi-player!", 8, 8, makecol(255, 255, 255));
		BLITBUF();
		clear_keybuf();
		readkey();
	}
}

int PlayGame(char *start_lvl)
{
	LEVEL *lvl, *next_lvl;
	int success = 0;

	lvl = LevelLoad(start_lvl);
	while(1)
	{
		if(lvl == NULL) return -1;

		success = RunLevel(lvl);
		if(success == 1)
		{
			next_lvl = LevelLoad(lvl->next_lvl);
			LevelFree(lvl);
			lvl = next_lvl;
			UnlockMulti();
		}
		else if(success == -1) // Player wants to quit!
		{
			LevelFree(lvl);
			return -1;
		}
	}

	LevelFree(lvl);
	return 0;
}

int RunLevel(LEVEL *lvl)
{
	int x, win;
	BITMAP *blvl, *dblvl;

	// Init
	for(x = 0; x < nplayers; x++)
	{
		players[x].x = lvl->startx;
		players[x].y = lvl->starty;
		players[x].health = 100;
		players[x].fireball.type = 0;
	}
	blvl = LevelRender(lvl);
	dblvl = create_bitmap(blvl->w, blvl->h);

	// Main Loop
	while(1)
	{
		if(InputCheck(lvl) == -1) return -1;
		Gravity(lvl);
		win = MoveObjects(lvl);
		RedrawLevel(lvl, blvl, dblvl);
		if(win == 1)
		{
			masked_blit(title_on, screen, 0, 56 * 6, (SCREEN_W - title_on->h) / 2, (SCREEN_H - 56) / 2, title_on->w, 56);
			clear_keybuf();
			readkey();
			return 1;
		}
	}

	destroy_bitmap(dblvl);
	destroy_bitmap(blvl);
}

void RedrawLevel(LEVEL *lvl, BITMAP *blvl, BITMAP *dblvl)
{
	int x, px, py;

	clear_bitmap(screenbuf);
	blit(blvl, dblvl, 0, 0, 0, 0, blvl->w, blvl->h);
	//blit(blvl, screenbuf, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

	LevelRenderObjects(lvl, dblvl);

	for(x = 0; x < nplayers; x++)
	{
		// Draw the player...
		masked_blit(graphics, dblvl, (players[x].sprite + 4) * 33, 0, players[x].x, players[x].y, 32, 32);
		// Draw their fireball...
		if(players[x].fireball.type == 1)
		{
			masked_blit(graphics, dblvl, 4 * 33, 4 * 33, players[x].fireball.x, players[x].fireball.y + 5 * sin(players[x].fireball.var1 * 3.0), 32, 32);
		}
	}

	for(x = 0; x < nplayers; x++)
	{
		// Get the "real" x/y
		px = players[x].x + 16;
		py = players[x].y + 16;
		// Move to viewport edge
		px = px - players[x].viewport[VIEWPORT_W] / 2;
		py = py - players[x].viewport[VIEWPORT_H] / 2;
		// Boundry check!
		if(px < 0) px = 0;
		else if(px + players[x].viewport[VIEWPORT_W] >= dblvl->w) px = dblvl->w - players[x].viewport[VIEWPORT_W];
		if(py < 0) py = 0;
		else if(py + players[x].viewport[VIEWPORT_H] >= dblvl->h) py = dblvl->h - players[x].viewport[VIEWPORT_H];
		// BLIT!
		blit(dblvl, screenbuf, px, py, players[x].viewport[VIEWPORT_X], players[x].viewport[VIEWPORT_Y], players[x].viewport[VIEWPORT_W], players[x].viewport[VIEWPORT_H]);
		rectfill(screenbuf, players[x].viewport[VIEWPORT_X], players[x].viewport[VIEWPORT_Y], players[x].viewport[VIEWPORT_X] + players[x].viewport[VIEWPORT_W] - 1, players[x].viewport[VIEWPORT_Y] + 16, makecol(0, 0, 0));
		textprintf(screenbuf, font, players[x].viewport[VIEWPORT_X] + 4, players[x].viewport[VIEWPORT_Y] + 4, makecol(255, 255, 255), "Lives: %d", players[x].lives);
		textout(screenbuf, font, "Health:", players[x].viewport[VIEWPORT_X] + 75, players[x].viewport[VIEWPORT_Y] + 4, makecol(255, 255, 255));
		rectfill(screenbuf, players[x].viewport[VIEWPORT_X] + 75 + text_length(font, "Health: "), players[x].viewport[VIEWPORT_Y] + 4, players[x].viewport[VIEWPORT_X] + text_length(font, "Health: ") + players[x].health * 2, players[x].viewport[VIEWPORT_Y] + 12, makecol(255, 0, 0));
	}

	BLITBUF();
}

int InputCheck(LEVEL *lvl)
{
	int x;

	if(key[KEY_ESC]) return -1;

	for(x = 0; x < nplayers; x++)
	{
		if(players[x].tick < 1) continue;
		if(key[players[x].keys[PKEY_JUMP]] && (!CanMoveHere(lvl, players[x].x, players[x].y + 1)))
		{
			players[x].fallrate = -10;
			players[x].tick = 0;
		}
		if(key[players[x].keys[PKEY_RIGHT]])
		{
			if(CanMoveHere(lvl, players[x].x + 4, players[x].y))
			{
				players[x].x += 4;
				players[x].sprite = players[x].sprite & (~0x02);
				players[x].sprite = players[x].sprite ^ 0x01;
				players[x].tick = 0;
			}
		}
		if(key[players[x].keys[PKEY_LEFT]])
		{
			if(CanMoveHere(lvl, players[x].x - 4, players[x].y))
			{
				players[x].x -= 4;
				players[x].sprite = players[x].sprite | 0x02;
				players[x].sprite = players[x].sprite ^ 0x01;
				players[x].tick = 0;
			}
		}
		if(key[players[x].keys[PKEY_SHOOT]] && players[x].fireball.type == 0)
		{
			players[x].fireball.type = 1 | (players[x].sprite & 0x02);
			players[x].fireball.x = players[x].x;
			players[x].fireball.y = players[x].y;
			players[x].fireball.var1 = 0;
			players[x].tick = 0;
		}
	}

	return 0;
}

void Gravity(LEVEL *lvl)
{
	int newpos, x;

	for(x = 0; x < nplayers; x++)
	{
		// Fall into the new position
		newpos = players[x].y + players[x].fallrate;

		// Are we in a tile?
		if(!CanMoveHere(lvl, players[x].x, newpos))
		{
			// Fine, move back up to the next tile, and stop falling...
			players[x].y = (newpos / 32 + (players[x].fallrate >= 0 ? 0 : 1)) * 32;
			players[x].fallrate = 0;
		}
		else
		{
			// Still falling. Increase fallrate, set new position.
			players[x].fallrate = (players[x].fallrate >= 8 ? players[x].fallrate : (players[x].fallrate + 1));
			players[x].y = newpos;
		}
	}

	return;
}

int CanMoveHere(LEVEL *lvl, int x, int y)
{
	int z;

	if(x < 0 || y < 0 || x + 32 > lvl->w * 32 || y + 32 > lvl->h * 32) return 0;

	if(IS_SOLID(GET_TILE(lvl, POS_TO_TILE(x), POS_TO_TILE(y))))
	{
		return 0;
	}

	if(x % 32 != 0)
	{
		if(IS_SOLID(GET_TILE(lvl, POS_TO_TILE(x) + 1, POS_TO_TILE(y))))
		{
			return 0;
		}
	}

	if(y % 32 != 0)
	{
		if(IS_SOLID(GET_TILE(lvl, POS_TO_TILE(x), POS_TO_TILE(y) + 1)))
		{
			return 0;
		}
	}

	if(x % 32 != 0 && y % 32 != 0)
	{
		if(IS_SOLID(GET_TILE(lvl, POS_TO_TILE(x) + 1, POS_TO_TILE(y) + 1)))
		{
			return 0;
		}
	}

	for(z = 0; z < lvl->nobjects; z++)
	{
		if(!(lvl->objects[z].type & 0x01)) continue;
		if(lvl->objects[z].x + 32 > x && lvl->objects[z].x - 32 < x && lvl->objects[z].y + 32 > y && lvl->objects[z].y - 32 < y)
		{
			return 0;
		}
	}

	return 1;
}

// Move the trolls and fireballs.
int MoveObjects(LEVEL *lvl)
{
	int x, y, newpos, d = 0;

	// Fireballs:
	for(x = 0; x < nplayers; x++)
	{
		if(players[x].fireball.type == 0) continue;
		newpos = players[x].fireball.type & 0x02 ? players[x].fireball.x - 8 : players[x].fireball.x + 8;
		if(!CanMoveHere(lvl, newpos, players[x].fireball.y))
		{
			players[x].fireball.type = 0;
			continue;
		}
		players[x].fireball.var1 += 8;
		if(players[x].fireball.var1 > 32 * 5)
		{
			players[x].fireball.type = 0;
		}
		else
		{
			players[x].fireball.x = newpos;
			// Check for a in-the-way troll:
			for(y = 0; y < lvl->nobjects; y++)
			{
				if(lvl->objects[y].type != 2) continue; // Not a troll.
				if(lvl->objects[y].x + 32 > players[x].fireball.x && lvl->objects[y].x - 32 < players[x].fireball.x && lvl->objects[y].y + 32 > players[x].fireball.y && lvl->objects[y].y - 32 < players[x].fireball.y)
				{
					// HIT!
					lvl->objects[y].var1 = 1;
					players[x].fireball.type = 0;
					if(AllTrollsDead(lvl)) d = 1;
				}
			}
		}
	}

	// Trolls:
	for(x = 0; x < lvl->nobjects; x++)
	{
		if(lvl->objects[x].type != 2 || (lvl->objects[x].type == 2 && lvl->objects[x].var1 != 0)) continue; // Not a living troll
		newpos = (lvl->objects[x].var2 == 0 ? (lvl->objects[x].x - 4) : (lvl->objects[x].x + 4));
		if(!CanMoveHere(lvl, newpos, lvl->objects[x].y) || CanMoveHere(lvl, newpos, lvl->objects[x].y + 1))
		{
			lvl->objects[x].var2 = lvl->objects[x].var2 ? 0 : 1;
		}
		else
		{
			lvl->objects[x].x = newpos;
		}
	}
	return d;
}

int AllTrollsDead(LEVEL *lvl)
{
	int x;

	for(x = 0; x < lvl->nobjects; x++)
	{
		if(lvl->objects[x].type != 2) continue;
		if(lvl->objects[x].var1 == 0) return 0;
	}

	return 1;
}

void UnlockMulti()
{
	FILE *fp;

	fp = fopen("multi.key", "wb");
	fclose(fp);
}
